class ReturnsDataReader:
    def __init__(self):
        pass

    def read_returns(self, assets, start_date, end_date):
        raise NotImplementedError("ReturnsDataReader.read_returns has not "
                                  "been implemented.")
