class FactorsDataReader:
    def __init__(self):
        pass

    def read_factors(self, start_date, end_date):
        raise NotImplementedError("FactorsDataReader.read_factors has not "
                                  "been implemented.")
